from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from imgui_bundle import imgui, imgui_ctx
from typing import Any, List, Set, Tuple
import webbrowser

Field = Tuple[str, str]

@dataclass
class Item:
  id     : int = 0
  summary: str = "this is summary"
  detail : str = "helloworld"
  fields : List[Field] = field(default_factory=list)
  feats  : Set[str]    = field(default_factory=set)

  def __post_init__(self):
    self.fields = [
      ("begin", "2024-01-24"),
      ("end", "2025-02-12"),
    ]
    self.feats = {"subtickets", "gantt"}


class IO(ABC):
  @abstractmethod
  def get_item(self) -> Item:
    pass

  @abstractmethod
  def set_summary(self, v: str) -> None:
    pass

  @abstractmethod
  def set_detail(self, v: str) -> None:
    pass

  @abstractmethod
  def set_field(self, idx: int, field: Field) -> None:
    pass

  @abstractmethod
  def delete_field(self, idx: int) -> None:
    pass

  @abstractmethod
  def reorder_field(self, before: int, after: int) -> None:
    pass


def compo(io: IO, item: Item = None) -> None:
  em = imgui.get_font_size()
  if item is None:
    item = io.get_item()

  with imgui_ctx.push_id(f"TicketEdit#{item.id}"):
    if imgui.collapsing_header("basic info", imgui.TreeNodeFlags_.default_open.value):
      changed, new_summary = imgui.input_text("summary", item.summary)
      if changed:
        io.set_summary(new_summary)

      changed, new_detail = imgui.input_text_multiline("detail", item.detail)
      if changed:
        io.set_detail(new_detail)

    if imgui.collapsing_header("fields", imgui.TreeNodeFlags_.default_open.value):
      count    = len(item.fields)
      drag_src = imgui.get_drag_drop_payload_py_id()
      if drag_src is not None and drag_src.type == "TicketEdit_Fields":
        drag_src = drag_src.data_id

      for idx in range(0, count+1):
        field = item.fields[idx] if idx < count else ("", "")

        with imgui_ctx.push_id(idx):
          with imgui_ctx.begin_group():
            w  = imgui.get_content_region_avail().x

            value_changed = False
            name_changed  = False

            imgui.begin_disabled(idx == count)
            imgui.text_link(":::")
            if imgui.begin_drag_drop_source():
              imgui.set_drag_drop_payload_py_id("TicketEdit_Fields", idx, imgui.Cond_.once.value);
              imgui.text(f"{field[0]}")
              imgui.end_drag_drop_source()
            if imgui.is_item_hovered():
              imgui.set_tooltip("right click to delete")
              if imgui.is_mouse_released(1):
                io.delete_field(idx)
            imgui.end_disabled()

            imgui.same_line()

            with imgui_ctx.push_item_width(6*em):
              name_changed, new_name = imgui.input_text_with_hint("##name", "name", field[0])

            imgui.same_line()

            with imgui_ctx.push_item_width(w - 10*em):
              value_changed, new_value = imgui.input_text_with_hint("##value", "value...", field[1])

            if name_changed or value_changed:
              io.set_field(idx, (new_name, new_value))

          if idx < count and imgui.begin_drag_drop_target():
            payload = imgui.accept_drag_drop_payload_py_id("TicketEdit_Fields")
            if payload is not None:
              io.reorder_field(payload.data_id, idx)
            imgui.end_drag_drop_target()

    if imgui.collapsing_header("features", imgui.TreeNodeFlags_.default_open.value):
      pass


def win(io: IO) -> None:
  item = io.get_item()

  with imgui_ctx.begin(f"#{item.id}: {item.summary}###TicketEdit#{item.id}"):
    compo(io, item)


class SimpleIO(IO):
  _item: Item = Item()

  def get_item(self) -> Item:
    return deepcopy(self._item)

  def set_summary(self, v: str) -> None:
    self._item.summary = v

  def set_detail(self, v: str) -> None:
    self._item.detail = v

  def set_field(self, idx: int, field: Field) -> None:
    if 0 <= idx and idx < len(self._item.fields):
      self._item.fields[idx] = field
    else:
      self._item.fields.append(field)

  def delete_field(self, idx: int) -> None:
    del self._item.fields[idx]

  def reorder_field(self, before: int, after: int) -> None:
    n = len(self._item.fields)
    if n == 0:
      return

    before = min(max(0, before), n-1)
    after  = min(max(0, after ), n-1)

    if before == after:
      return

    temp = self._item.fields[before]
    del self._item.fields[before]
    self._item.fields.insert(after, temp)
